import time
import logging
import pigpio
from .ADS1256_definitions import *
from . import ADS1256_default_config

logger = logging.getLogger("PiPyADC")

class ADS1256():
    """Python class for interfacing the ADS1256 and ADS1255 analog to
    digital converters with the Raspberry Pi.

    This is part of module PiPyADC
    Download: https://github.com/ul-gh/PiPyADC
    
    Default pin and settings configuration is for the Open Hardware
    "Waveshare High-Precision AD/DA Board"

    See file ADS1256_default_config.py for
    configuration settings and description.

    Register read/write access is implemented via Python class/instance
    properties. Commands are implemented as functions.

    See help(ADS1256) for usage of the properties and functions herein.

    See ADS1256_definitions.py for chip registers, flags and commands.
    
    Documentation source: Texas Instruments ADS1255/ADS1256
    datasheet SBAS288: http://www.ti.com/lit/ds/sbas288j/sbas288j.pdf
    """
    open_spi_handles = {}
    pins_initialized = {}
    exclusive_pins_used = {}

    # Hardware pin configuration must be set at initialization phase.
    # Register/Configuration Flag settings are initialized, but these
    # can be changed during runtime via class properties.
    # Default config is read from external file (module).
    def __init__(self, conf=ADS1256_default_config, pi=None):
        # Set up the pigpio object if not provided as an argument
        if pi is None:
            self.pi = pigpio.pi()
            print(self.pi)
            self.created_pigpio = True
        else:
            self.pi = pi
            self.created_pigpio = False
        if not self.pi.connected:
            raise IOError("Could not connect to hardware via pigpio library")
        # This is not needed for any function currently implemented here,
        # but the attribute is kept for user code which relys on the value
        self.v_ref = conf.v_ref
        # Following constants are for the bit-banging funcionality and for
        # timing specifics of the ADS1256 chip
        self._DRDY_TIMEOUT = conf.DRDY_TIMEOUT
        self._DRDY_DELAY = conf.DRDY_DELAY
        # ADS1255/ADS1256 command timing specifications.
        # Delay between requesting data and reading the bus for
        # RDATA, RDATAC and RREG commands (datasheet: t_6 >= 50*CLKIN period).
        self._DATA_TIMEOUT = 1E-6 + 50/conf.CLKIN_FREQUENCY
        # Command-to-command timeout after SYNC and RDATAC
        # commands (datasheet: t11)
        self._SYNC_TIMEOUT = 1E-6 + 24/conf.CLKIN_FREQUENCY
        # See datasheet ADS1256: CS needs to remain low
        # for t_10 = 8*T_CLKIN after last SCLK falling edge of a command.
        # Because this delay is longer than timeout t_11 for the
        # RREG, WREG and RDATA commands of 4*T_CLKIN, we do not need
        # the extra t_11 timeout for these commands when using software
        # chip select selection and the _CS_TIMEOUT.
        self._CS_TIMEOUT = 1E-6 + 8/conf.CLKIN_FREQUENCY
        # When using hardware/hard-wired chip select, still a command-
        # to command timeout of t_11 is needed as a minimum for the
        # RREG, WREG and RDATA commands.
        self._T_11_TIMEOUT = 1E-6 + 4/conf.CLKIN_FREQUENCY
        self.configure_gpios(conf)
        self.configure_spi(conf)
        # Device reset for defined initial state
        if conf.CHIP_HARD_RESET_ON_START:
            self.hard_reset()
        else:
            self.reset()
        self.preset_adc_registers(conf)
        self.check_chip_id()

    def __enter__(self):
        return self

    def __exit__(self, _exc_type, _exc_value, _traceback):
        if _exc_type is not None:
            self.stop_close_all()
        else:
            self.stop()


    def configure_gpios(self, conf):
        # The following four GPIOs are used for this ADS1256 implementation:
        self._CS_PIN = conf.CS_PIN
        self._DRDY_PIN = conf.DRDY_PIN
        self._RESET_PIN = conf.RESET_PIN
        # Not implemented
        # self._PDWN_PIN = conf.PDWN_PIN
        # Config and initialize the SPI and GPIO pins used by the ADC.
        # For configuration of multiple SPI devices on this bus:
        if conf.CS_PIN in self.exclusive_pins_used:
            self.stop_close_all()
            raise RuntimeError("CS pin already used. Must be exclusive!")
        self.exclusive_pins_used[conf.CS_PIN] = conf.CS_PIN
        if hasattr(conf, "CHIP_SELECT_GPIOS_INITIALIZE"):
            if type(conf.CHIP_SELECT_GPIOS_INITIALIZE) is int:
                cs_gpios = (conf.CHIP_SELECT_GPIOS_INITIALIZE, )
            else:
                cs_gpios = tuple(conf.CHIP_SELECT_GPIOS_INITIALIZE)
            if conf.CS_PIN not in cs_gpios:
                cs_gpios = cs_gpios + (conf.CS_PIN, )
            for pin in cs_gpios:
                self._init_input(pin, init_state=1, name="chip select")
        if conf.DRDY_PIN in self.exclusive_pins_used:
            self.stop_close_all()
            raise RuntimeError("DRDY pin already used. Must be exclusive!")
        # DRDY_PIN is the only GPIO input used by this ADC
        if conf.DRDY_PIN is not None:
            self._DRDY_PIN = conf.DRDY_PIN
            self.pi.set_mode(conf.DRDY_PIN, pigpio.INPUT)
            self.exclusive_pins_used[conf.DRDY_PIN] = conf.DRDY_PIN
        # GPIO Outputs. If chip select pin is set to None, the
        # respective ADC input pin is assumed to be hardwired to GND.
        self._init_output(conf.RESET_PIN, init_state=1, name="reset pin")
        self._init_output(conf.PDWN_PIN, init_state=1, name="pdwn pin")
    
    def configure_spi(self, conf):
        # SPI bus config
        logger.debug(f"Activating SPI, SW chip select on GPIO: {conf.CS_PIN}")
        # The ADS1256 uses SPI MODE=1 <=> CPOL=0, CPHA=1. 
        if hasattr(conf, "SPI_FLAGS"):
            spi_flags = conf.SPI_FLAGS
        else:
            #              bbbbbbRTnnnnWAuuupppmm
            spi_flags  = 0b0000000000000011100001 
        if hasattr(conf, "SPI_BUS") and conf.SPI_BUS == 1:
            spi_flags |= 0b0000000000000100000000 
        # PIGPIO library returns a numeric handle for each chip on this bus.
        try:
            self.spi_handle = self.pi.spi_open(0, conf.SPI_FREQUENCY, spi_flags)
        except Exception as e:
            logger.error("SPI open error")
            self.stop_close_all()
            raise e from None
        # Add to class attribute
        self.open_spi_handles[self.spi_handle] = self.spi_handle
        logger.debug(f"Obtained SPI device handle: {self.spi_handle}")
        
    def preset_adc_registers(self, conf):
        # Configure ADC registers:
        # Status register not yet set, only variable written to avoid multiple
        # triggering of the AUTOCAL procedure by changing other register flags
        self._status       = conf.status
        # Class properties now configure registers via their setter functions
        self.mux           = conf.mux
        self.adcon         = conf.adcon
        self.drate         = conf.drate
        self.gpio          = conf.gpio
        self.status        = conf.status

    def check_chip_id(self):
        # This invokes a getter function..
        chip_ID = self.chip_ID
        logger.debug(f"Chip ID: {chip_ID}")
        if chip_ID != 3:
            self.stop_close_all()
            raise RuntimeError("Received wrong chip ID value for ADS1256. "
                               "Hardware connected?")

    def stop(self):
        """Close own SPI handle, only stop pigpio connection if we created it
        """
        logger.debug(f"Closing SPI handle: {self.spi_handle}")
        self.pi.spi_close(self.spi_handle)
        self.open_spi_handles.pop(self.spi_handle)
        self.exclusive_pins_used.pop(self._CS_PIN)
        self.exclusive_pins_used.pop(self._DRDY_PIN)
        if self.created_pigpio:
            logger.debug(f"Closing PIGPIO instance")
            self.pi.stop()
        # Else leaving external PIGPIO instance active

    def stop_close_all(self):
        """Close all open pigpio SPI handles and stop pigpio connection
        """
        for handle in reversed(self.open_spi_handles):
            logger.debug(f"Closing SPI handle: {handle}")
            self.pi.spi_close(handle)
        self.open_spi_handles.clear()
        self.exclusive_pins_used.clear()
        logger.debug(f"Closing PIGPIO instance")
        self.pi.stop()


    @property
    def pga_gain(self):
        """Get/Set ADC programmable gain amplifier setting.
        
        The available options for the ADS1256 are:
        1, 2, 4, 8, 16, 32 and 64.

        This function sets the ADCON register with the code values
        defined in file ADS1256_definitions.py.

        Note: When changing the gain setting at runtime, with activated
        ACAL flag (AUTOCAL_ENABLE), this causes a _wait_DRDY() timeout
        for the calibration process to finish.
        """
        return 2**(self.read_reg(REG_ADCON)&0b111)
    @pga_gain.setter
    def pga_gain(self, value):
        if value not in (1, 2, 4, 8, 16, 32, 64):
            self.stop_close_all()
            raise ValueError("Argument must be one of: 1, 2, 4, 8, 16, 32, 64")
        else:
            log2val = int.bit_length(value) - 1
            self.write_reg(REG_ADCON, self.adcon&0b11111000 | log2val)
            if self._status & AUTOCAL_ENABLE:
                self._wait_DRDY()

    @property
    def v_per_digit(self):
        """Get ADC LSB weight in volts per numeric output digit.
        Readonly: This is a convenience value calculated from
        gain and v_ref setting.
        """
        return self.v_ref * 2.0/(self.pga_gain * (2**23 - 1))
    @v_per_digit.setter
    def v_per_digit(self, value):
        self.stop_close_all()
        raise AttributeError("This is a read-only attribute")

    @property
    def status(self):
        """Get/Set value of ADC status register, REG_STATUS (8 bit).
        For available settings flag options, see datasheet and file
        ADS1256_definitions.py. Note: When enabling the AUTOCAL
        flag, any subsequent access to the BUFEN flag, DRATE register
        (drate property) or PGA gain setting (gain property) will cause
        an additional delay for completion of hardware auto-calibration.
        """
        return self.read_reg(REG_STATUS)
    @status.setter
    def status(self, value):
        self.write_reg(REG_STATUS, value)
        self._status = value
        # When AUTOCAL flag has been enabled, a _wait_DRDY() is needed here.
        # When AUTOCAL flag had been enabled before and the BUFEN flag has not
        # been changed, this is likely not necessary, but FIXME: Better risk
        # a small possibly unnecessary delay than risk invalid data or settings.
        # Thus, if AUTOCAL flag /is/ enabled, always do a _wait_DRDY().
        if self._status & AUTOCAL_ENABLE:
            self._wait_DRDY()

    @property
    def mux(self):
        """Get/Set value of ADC analog input multiplexer register,
        REG_MUX, used for selecting any arbitrary pair of input pins
        as a differential input channel. For single-ended measurements,
        choose NEG_AINCOM as the second input pin.

        The most significant four bits select the positive input pin.
        The least significant four bits select the negative input pin.
        
        Example: ads1256.mux = POS_AIN4 | NEG_AINCOM.

        IMPORTANT:

        When switching inputs during a running conversion cycle,
        invalid data is acquired.

        To prevent this, you must restart the conversion using the
        sync() function or the SYNC hardware pin before doing an
        async_read().

        The resulting delay can be avoided. See functions:

        read_continue()
            for cyclic reads of multiple channels at once - the ADC must not be
            reconfigured between invocations of this function, otherwise false
            data is read for the first value of the sequence

        read_sequence()
            for reading a succession of multiple channels at once, configuring
            all input channels including the first one for each cycle
        """
        return self.read_reg(REG_MUX)
    @mux.setter
    def mux(self, value):
        self.write_reg(REG_MUX, value)

    @property
    def adcon(self):
        """Get/Set value of the ADC configuration register, REG_ADCON.
        Note: When the AUTOCAL flag is enabled, this causes a
        _wait_DRDY() timeout.
        """
        return self.read_reg(REG_ADCON) 
    @adcon.setter
    def adcon(self, value):
        self.write_reg(REG_ADCON, value)
        if self._status & AUTOCAL_ENABLE:
            self._wait_DRDY()

    @property
    def drate(self):
        """Get/Set value of the ADC output sample data rate by setting
        the DRATE register (REG_DRATE).
        This configures the hardware integrated moving average filter.

        When changing the register during a running acquisition,
        invalid data is sampled. In this case, call the sync() method
        to restart the acquisition cycle.

        The available data rates are defined in ADS1256_definitions.py.
        """
        return self.read_reg(REG_DRATE)
    @drate.setter
    def drate(self, value):
        self.write_reg(REG_DRATE, value)
        if self._status & AUTOCAL_ENABLE:
            self._wait_DRDY()

    @property
    def gpio(self):
        """Get the logic level of the four GPIO pins, returned as
        a four-bit bitmask or Set the status of the GPIO register,
        REG_IO, where the most significant four bits represent the
        pin direction, and the least significant four bits determine
        the output logic level.
        A timeout/debounce for the reading is not implemented.
        """
        return 0x0F & self.read_reg(REG_IO)
    @gpio.setter
    def gpio(self, value):
        self.write_reg(REG_IO, value)

    @property
    def ofc(self):
        """Get/Set the three offset compensation registers, OFC0..2.
        This property is supposed to be a signed integer value.
        Gets/sets 24-bit two's complement value in three 8-bit-registers.
        """
        ofc0 = self.read_reg(REG_OFC0)
        ofc1 = self.read_reg(REG_OFC1)
        ofc2 = self.read_reg(REG_OFC2)
        int24_result = ofc2<<16 | ofc1<<8 | ofc0
        # Take care of 24-Bit 2's complement.
        if int24_result < 0x800000:
            return int24_result
        else:
            return int24_result - 0x1000000
    @ofc.setter
    def ofc(self, value):
        value = int(value)
        if value < -0x800000 or value > 0x7FFFFF:
            self.stop_close_all()
            raise ValueError("Error: Offset value out of signed int24 range")
        else:
            # Generate 24-Bit 2's complement.
            if value < 0:
                value += 0x1000000
            # self._send_byte() automatically truncates to uint8
            self.write_reg(REG_OFC0, value)
            value >>= 8
            self.write_reg(REG_OFC1, value)
            value >>= 8
            self.write_reg(REG_OFC2, value)
 
    @property
    def fsc(self):
        """Get/Set the three full-scale adjustment registers, OFC0..2.
        This property is supposed to be a positive integer value.
        Gets/sets 24-bit unsigned int value in three 8-bit-registers.
        """
        fsc0 = self.read_reg(REG_FSC0)
        fsc1 = self.read_reg(REG_FSC1)
        fsc2 = self.read_reg(REG_FSC2)
        return fsc2<<16 | fsc1<<8 | fsc0
    @fsc.setter
    def fsc(self, value):
        value = int(value)
        if value < 0 or value > 0xFFFFFF:
            self.stop_close_all()
            raise ValueError("Error: This must be a positive int in 24-bit range")
        else:
            # self._send_byte() automatically truncates to uint8
            self.write_reg(REG_FSC0, value)
            value >>= 8
            self.write_reg(REG_FSC1, value)
            value >>= 8
            self.write_reg(REG_FSC2, value)

    @property
    def chip_ID(self):
        """Get the numeric ID from the ADS chip.
        Useful to check if hardware is connected.

        Value for the ADS1256 on the Waveshare board seems to be a 3.
        """
        self._wait_DRDY()
        return self.read_reg(REG_STATUS) >> 4
    @chip_ID.setter
    def chip_ID(self, value):
        self.stop_close_all()
        raise AttributeError("This is a read-only attribute")

    def wait_DRDY(self):
        """wait_DRDY() is now obsolete as the DRDY handling takes place
        automatically when calling the STATUS, ADCON and DRATE register setters
        """
        logger.warn("wait_DRDY is now an obsolete function.")
        self._wait_DRDY()

    def read_reg(self, register):
        """Returns data byte from the specified register
        
        Argument: register address
        """
        self._chip_select()
        self.pi.spi_write(self.spi_handle, [CMD_RREG|register, 0x00])
        time.sleep(self._DATA_TIMEOUT)
        count, inbytes = self.pi.spi_read(self.spi_handle, 1)
        if count != 1:
            logger.error("SPI read error occurred!")
        # Release chip select and implement t_11 timeout
        self._chip_release()
        return inbytes[0]


    def write_reg(self, register, data):
        """Writes data byte to the specified register
 
        Arguments: register address, data byte (uint_8)
        """
        self._chip_select()
        self.pi.spi_write(self.spi_handle, [CMD_WREG|register, 0x00, data])
        # Release chip select and implement t_11 timeout
        self._chip_release()


    def cal_self_offset(self):
        """Perform an input zero calibration using chip-internal
        reference switches.

        Sets the ADS1255/ADS1256 OFC register.
        """
        self._chip_select()
        self.pi.spi_write(self.spi_handle, [CMD_SELFOCAL])
        self._wait_DRDY()
        # Release chip select and implement t_11 timeout
        self._chip_release()


    def cal_self_gain(self):
        """Perform an input full-scale calibration
        using chip-internal circuitry connected to VREFP and VREFN.

        Sets the ADS1255/ADS1256 FSC register.
        """
        self._chip_select()
        self.pi.spi_write(self.spi_handle, [CMD_SELFGCAL])
        self._wait_DRDY()
        # Release chip select and implement t_11 timeout
        self._chip_release()


    def cal_self(self):
        """Perform an input zero and full-scale two-point-calibration
        using chip-internal circuitry connected to VREFP and VREFN.

        Sets the ADS1255/ADS1256 OFC and FSC registers.
        """
        self._chip_select()
        self.pi.spi_write(self.spi_handle, [CMD_SELFCAL])
        self._wait_DRDY()
        # Release chip select and implement t_11 timeout
        self._chip_release()


    def cal_system_offset(self):
        """Set the ADS1255/ADS1256 OFC register such that the
        current input voltage corresponds to a zero output value.
        The input multiplexer must be set to the appropriate pins first.
        """
        self._chip_select()
        self.pi.spi_write(self.spi_handle, [CMD_SYSOCAL])
        self._wait_DRDY()
        # Release chip select and implement t_11 timeout
        self._chip_release()


    def cal_system_gain(self):
        """Set the ADS1255/ADS1256 FSC register such that the current
        input voltage corresponds to a full-scale output value.
        The input multiplexer must be set to the appropriate pins first.
        """
        self._chip_select()
        self.pi.spi_write(self.spi_handle, [CMD_SYSGCAL])
        self._wait_DRDY()
        # Release chip select and implement t_11 timeout
        self._chip_release()


    def standby(self):
        """Put chip in low-power standby mode
        """
        self._chip_select()
        self.pi.spi_write(self.spi_handle, [CMD_STANDBY])
        # Release chip select and implement t_11 timeout
        self._chip_release()


    def wakeup(self):
        """Wake up the chip from standby mode.
        See datasheet for settling time specifications after wake-up.
        Data is ready when the DRDY pin becomes active low.

        You can then use the read_oneshot() function to read a new
        sample of input data.

        Call standby() to enter standby mode again.
        """
        self._chip_select()
        self.pi.spi_write(self.spi_handle, [CMD_WAKEUP])
        # Release chip select and implement t_11 timeout
        self._chip_release()

    def reset(self):
        """Reset all registers except CLK0 and CLK1 bits
        to reset values and Polls for DRDY change / timeout afterwards.
        """
        self._chip_select()
        self.pi.spi_write(self.spi_handle, [CMD_RESET])
        self._wait_DRDY()
        # Release chip select and implement t_11 timeout
        self._chip_release()

    def hard_reset(self):
        """Reset by toggling the hardware pin as configured as "RESET_PIN".
        """
        if self._RESET_PIN is None:
            self.stop_close_all()
            raise RuntimeError("Reset pin is not configured!")
        else:
            logger.debug("Performing hard RESET...")
            self.pi.write(self._RESET_PIN, 0)
            time.sleep(100E-6)
            self.pi.write(self._RESET_PIN, 1)
            # At hardware initialisation, a settling time for the oscillator
            # is necessary before doing any register access.
            # This is approx. 30ms, according to the datasheet.
            time.sleep(0.03)
            self._wait_DRDY()
 

    def sync(self):
        """Restart the ADC conversion cycle with a SYNC + WAKEUP
        command sequence as described in the ADS1256 datasheet.
        
        This is useful to restart the acquisition cycle after rapid
        changes of the input signals, for example when using an
        external input multiplexer or after changing ADC configuration
        flags.
        """
        self._chip_select()
        self.pi.spi_write(self.spi_handle, [CMD_SYNC])
        time.sleep(self._SYNC_TIMEOUT)
        self.pi.spi_write(self.spi_handle, [CMD_WAKEUP])
        # Release chip select and implement t_11 timeout
        self._chip_release()


    def read_async(self):
        """Read ADC result as soon as possible
        
        Arguments:  None
        Returns:    Signed integer ADC conversion result

        Issue this command to read a single conversion result for a
        previously set /and stable/ input channel configuration.

        For the default, free-running mode of the ADC, this means
        invalid data is returned when not synchronising acquisition
        and input channel configuration changes.

        To avoid this, after changing input channel configuration or
        with an external hardware multiplexer, use the hardware SYNC
        input pin or use the sync() method to restart the
        conversion cycle before calling read_async().
        
        Because this function does not implicitly restart a running
        acquisition, it is faster that the read_oneshot() method.
        """
        self._chip_select()
        # Wait for data to be ready
        self._wait_DRDY()
        # Send the read command
        self.pi.spi_write(self.spi_handle, [CMD_RDATA])
        time.sleep(self._DATA_TIMEOUT)
        # The result is 24 bits little endian two's complement value by default
        count, inbytes = self.pi.spi_read(self.spi_handle, 3)
        if count != 3:
            logger.error("SPI read error occurred!")
        self._chip_release()
        return int.from_bytes(inbytes, "big", signed=True)


    def read_oneshot(self, diff_channel):
        """Restart/re-sync ADC and read the specified input pin pair.
        
        Arguments:  8-bit code value for differential input channel
                        (See definitions for the REG_MUX register)
        Returns:    Signed integer conversion result

        Use this function after waking up from STANDBY mode.
        
        When switching inputs during a running conversion cycle,
        invalid data is acquired.

        To prevent this, this function automatically restarts the
        conversion cycle after configuring the input channels.

        The resulting delay can be avoided. See functions:

        read_continue()
            for cyclic reads of multiple channels at once - the ADC must not be
            reconfigured between invocations of this function, otherwise false
            data is read for the first value of the sequence

        read_sequence()
            for reading a succession of multiple channels at once, configuring
            all input channels including the first one for each cycle
        """
        self._chip_select()
        # Set input pin mux position for this cycle"
        self.pi.spi_write(
            self.spi_handle, [CMD_WREG|REG_MUX, 0x00, diff_channel, CMD_SYNC])
        time.sleep(self._SYNC_TIMEOUT)
        self.pi.spi_write(self.spi_handle, [CMD_WAKEUP])
        self._wait_DRDY()
        # Read data from ADC, which still returns the /previous/ conversion
        # result from before changing inputs
        self.pi.spi_write(self.spi_handle, [CMD_RDATA])
        time.sleep(self._DATA_TIMEOUT)
        # The result is 24 bits little endian two's complement value by default
        count, inbytes = self.pi.spi_read(self.spi_handle, 3)
        if count != 3:
            logger.error("SPI read error occurred!")
        self._chip_release()
        return int.from_bytes(inbytes, "big", signed=True)

    def read_and_next_is(self, diff_channel):
        """Reads ADC data of presently running or already finished
        conversion, sets and synchronises new input channel config
        for next sequential read.

        Arguments:  8-bit code value for differential input channel
                        (See definitions for the REG_MUX register)
        Returns:    Signed integer conversion result for present read
        
        This enables rapid dycling through different channels and
        implements the timing sequence outlined in the ADS1256
        datasheet (Sept.2013) on page 21, figure 19: "Cycling the
        ADS1256 Input Multiplexer".

        Note: In most cases, a fixed sequence of input channels is known
        beforehand. For that case, this module implements the function:
        
        read_sequence(ch_sequence)
            which automates the process for cyclic data acquisition.
        """
        self._chip_select()
        self._wait_DRDY()
        # Setting mux position for next cycle"
        self.pi.spi_write(
            self.spi_handle, [CMD_WREG|REG_MUX, 0x00, diff_channel, CMD_SYNC])
        time.sleep(self._SYNC_TIMEOUT)
        self.pi.spi_write(self.spi_handle, [CMD_WAKEUP])
        # The datasheet is a bit unclear if a t_11 timeout is needed here.
        # Assuming the extra timeout is the safe choice:
        time.sleep(self._T_11_TIMEOUT)
        # Read data from ADC, which still returns the /previous/ conversion
        # result from before changing inputs
        self.pi.spi_write(self.spi_handle, [CMD_RDATA])
        time.sleep(self._DATA_TIMEOUT)
        # The result is 24 bits little endian two's complement value by default
        count, inbytes = self.pi.spi_read(self.spi_handle, 3)
        if count != 3:
            logger.error("SPI read error occurred!")
        self._chip_release()
        return int.from_bytes(inbytes, "big", signed=True)

    def read_continue(self, ch_sequence, ch_buffer=None):
        """Continues reading a cyclic sequence of ADC input channel pin pairs.

        The first data sample is only valid if the ADC data register contains
        valid data from a previous conversion. I.e. the last element of the
        ch_sequence must be the first channel configuration to be read during
        the next following cycle.

        For short sequences, this is faster than the read_sequence() method
        because it does not interrupt an already running and pre-configured
        conversion cycle.

        Argument1:  Tuple (list) of 8-bit code values for differential
                    input channel pins to read sequentially in a cycle.
                    (See definitions for the REG_MUX register)

                    Example:
                    ch_sequence=(POS_AIN0|NEG_AIN1, POS_AIN2|NEG_AINCOM)

        Argument2:  List (array, buffer) of signed integer conversion
                    results for the sequence of input channels.

        Returns:    List (array, buffer) of signed integer conversion
                    results for the sequence of input channels.

        This implements the timing sequence outlined in the ADS1256
        datasheet (Sept.2013) on page 21, figure 19: "Cycling the
        ADS1256 Input Multiplexer" for cyclic data acquisition.
        """
        buf_len = len(ch_sequence)
        if ch_buffer is None:
            ch_buffer = [0] * buf_len
        for i in range(0, buf_len):
            ch_buffer[i] = self.read_and_next_is(ch_sequence[(i+1)%buf_len])
        return ch_buffer

    def read_sequence(self, ch_sequence, ch_buffer=None):
        """Reads a sequence of ADC input channel pin pairs.

        Restarts and re-syncs the ADC for the first sample.

        The time delay resulting from this can be avoided when reading
        the ADC in a cyclical pattern using the read_continue() method.

        Argument1:  Tuple (list) of 8-bit code values for differential
                    input channel pins to read sequentially in a cycle.
                    (See definitions for the REG_MUX register)

                    Example:
                    ch_sequence=(POS_AIN0|NEG_AIN1, POS_AIN2|NEG_AINCOM)

        Argument2:  List (array, buffer) of signed integer conversion
                    results for the sequence of input channels.

        Returns:    List (array, buffer) of signed integer conversion
                    results for the sequence of input channels.

        This implements the timing sequence outlined in the ADS1256
        datasheet (Sept.2013) on page 21, figure 19: "Cycling the
        ADS1256 Input Multiplexer" for cyclic data acquisition.
        """
        self.mux = ch_sequence[0]
        self.sync()
        buf_len = len(ch_sequence)
        if ch_buffer is None:
            ch_buffer = [0] * buf_len
        for i in range(0, buf_len):
            ch_buffer[i] = self.read_and_next_is(ch_sequence[(i+1)%buf_len])
        return ch_buffer


    # Delays until the configured DRDY input pin is pulled to
    # active logic low level by the ADS1256 hardware or until the
    # _DRDY_TIMEOUT has passed.
    # 
    # The minimum necessary _DRDY_TIMEOUT when not using the hardware
    # pin, can be up to approx. one and a half second, see datasheet..
    # 
    # This delay is also necessary when using the automatic calibration feature
    # (ACAL flag), after every access that changes the PGA gain bits in
    # ADCON register, the DRATE register or the BUFFEN flag in status register.
    def _wait_DRDY(self):
        start = time.time()
        elapsed = time.time() - start
        # Waits for DRDY pin to go to active low or _DRDY_TIMEOUT seconds to pass
        if self._DRDY_PIN is not None:
            drdy_level = self.pi.read(self._DRDY_PIN)
            while (drdy_level == 1) and (elapsed < self._DRDY_TIMEOUT):
                elapsed = time.time() - start
                drdy_level = self.pi.read(self._DRDY_PIN)
                # Sleep in order to avoid busy wait and reduce CPU load.
                time.sleep(self._DRDY_DELAY)
            if elapsed >= self._DRDY_TIMEOUT:
                logger.warning("Timeout while polling configured DRDY pin!")
        else:
            time.sleep(self._DRDY_TIMEOUT)

    def _chip_select(self):
        # If chip select pin is hardwired to GND, do nothing.
        if self._CS_PIN is not None:
            self.pi.write(self._CS_PIN, 0)

    # Release chip select and implement t_11 timeout
    def _chip_release(self):
        if self._CS_PIN is not None:
            time.sleep(self._CS_TIMEOUT)
            self.pi.write(self._CS_PIN, 1)
        else:
            # The minimum t_11 timeout between commands, see datasheet Figure 1.
            time.sleep(self._T_11_TIMEOUT)

    def _init_output(self, pin, init_state, name="output"):
        if pin is not None and pin not in self.pins_initialized:
            logger.debug(f"Setting as output: {pin} ({name})")
            self.pi.set_mode(pin, pigpio.OUTPUT)
            self.pi.write(pin, init_state)
            self.pins_initialized[pin] = pin

    def _init_input(self, pin, init_state, name="input"):
        if pin is not None and pin not in self.pins_initialized:
            logger.debug(f"Setting as input: {pin} ({name})")
            self.pi.set_mode(pin, pigpio.INPUT)
            self.pi.write(pin, init_state)
            self.pins_initialized[pin] = pin



class TextScreen():
    """Output arbitrary number of lines of text, while re-using the screen
    area previously written to.

    For this, text is put in a buffer using the put() method, and output is
    written to the screen all at once when the refresh() method is called.

    Ulrich Lukas 2022-08-27
    """
    def __init__(self):
        self._lines_in_buffer = 0
        self._lines_printed = 0
        self._text_buffer = ""

    def put(self, text):
        """Put lines of text in buffer, but do not output anything.

        Write output all at once, clearing the previous screen contents,
        when refresh() is later called.
        """
        self._lines_in_buffer += 1 + text.count("\n")
        self._text_buffer += text.replace("\n", "\x1B[0K\n") + "\x1B[0K\n"

    def refresh(self):
        """Write output from buffer to screen, clearing previous content
        """
        clear_code = f"\x1B[{self._lines_printed}F" if self._lines_printed else ""
        print(f"{clear_code}{self._text_buffer}", end="")
        self._lines_printed = self._lines_in_buffer
        self._text_buffer = ""
        self._lines_in_buffer = 0





################### CONSTANT DEFINITIONS for class ADS1256 ####################
# Register addresses:
REG_STATUS  = 0x00
REG_MUX     = 0x01
REG_ADCON   = 0x02
REG_DRATE   = 0x03
REG_IO      = 0x04
REG_OFC0    = 0x05
REG_OFC1    = 0x06
REG_OFC2    = 0x07
REG_FSC0    = 0x08
REG_FSC1    = 0x09
REG_FSC2    = 0x0A
NUM_REG     = 11

# REG_STATUS: Flags:
BUFFER_ENABLE    = 0x02
AUTOCAL_ENABLE   = 0x04
ORDER_LSB        = 0x08

# REG_MUX: High nibble selects positive input, low nibble negative input.
# Pin selection codes: logic OR together to form the register value.
# Example: ads1256.mux = POS_AIN0 | NEG_AINCOM
# Pin selection codes for the positive input:
POS_AIN0   = 0x00
POS_AIN1   = 0x10
POS_AIN2   = 0x20
POS_AIN3   = 0x30
POS_AIN4   = 0x40
POS_AIN5   = 0x50
POS_AIN6   = 0x60
POS_AIN7   = 0x70
POS_AINCOM = 0x80
# Pin selection codes for the negative input:
NEG_AIN0   = 0x00
NEG_AIN1   = 0x01
NEG_AIN2   = 0x02
NEG_AIN3   = 0x03
NEG_AIN4   = 0x04
NEG_AIN5   = 0x05
NEG_AIN6   = 0x06
NEG_AIN7   = 0x07
NEG_AINCOM = 0x08

# REG_ADCON: Gain levels
# Note: You can set the numeric values 1, 2, 4, 8, 16, 32 and 64 directly
# via class/instance property ADS1256.pga_gain.
GAIN_1      = 0x00
GAIN_2      = 0x01
GAIN_4      = 0x02
GAIN_8      = 0x03
GAIN_16     = 0x04
GAIN_32     = 0x05
GAIN_64     = 0x06

# REG_ADCON: Sensor Detect Current Sources
SDCS_OFF     = 0x00
SDCS_500pA   = 0x08
SDCS_2uA     = 0x10
SDCS_10uA    = 0x18

# REG_ADCON: Clock output pin settings
CLKOUT_OFF      = 0x00
CLKOUT_EQUAL    = 0x20
CLKOUT_HALF     = 0x40
CLKOUT_FOURTH   = 0x60

# REG_DRATE: Sample rate definitions:
DRATE_30000     = 0b11110000 # 30,000SPS (default)
DRATE_15000     = 0b11100000 # 15,000SPS
DRATE_7500      = 0b11010000 # 7,500SPS
DRATE_3750      = 0b11000000 # 3,750SPS
DRATE_2000      = 0b10110000 # 2,000SPS
DRATE_1000      = 0b10100001 # 1,000SPS
DRATE_500       = 0b10010010 # 500SPS
DRATE_100       = 0b10000010 # 100SPS
DRATE_60        = 0b01110010 # 60SPS
DRATE_50        = 0b01100011 # 50SPS
DRATE_30        = 0b01010011 # 30SPS
DRATE_25        = 0b01000011 # 25SPS
DRATE_15        = 0b00110011 # 15SPS
DRATE_10        = 0b00100011 # 10SPS
DRATE_5         = 0b00010011 # 5SPS
DRATE_2_5       = 0b00000011 # 2.5SPS


# Chip-level-command Definitions. Special functions implemented in this
# class take care of command order / timeouts and handling of data/results.
CMD_WAKEUP  = 0x00 # Completes SYNC and exits standby mode
CMD_RDATA   = 0x01 # Read data
CMD_RDATAC  = 0x03 # Start read data continuously
CMD_SDATAC  = 0x0F # Stop read data continuously
CMD_RREG    = 0x10 # Read from register
CMD_WREG    = 0x50 # Write to register
CMD_SELFCAL = 0xF0 # Offset and gain self-calibration
CMD_SELFOCAL= 0xF1 # Offset self-calibration
CMD_SELFGCAL= 0xF2 # Gain self-calibration
CMD_SYSOCAL = 0xF3 # System offset calibration
CMD_SYSGCAL = 0xF4 # System gain calibration
CMD_SYNC    = 0xFC # Synchronize the A/D conversion
CMD_STANDBY = 0xFD # Begin standby mode
CMD_RESET   = 0xFE # Reset to power-on values



import logging
from pipyadc.ADS1256_definitions import *
################ Configuration file for one ADS1256 instance  #################
# SPI bus configuration and GPIO pins used for the ADS1255/ADS1256.
#
# These settings are compatible with the "Waveshare High-Precision AD/DA" board.
#
# To create multiple class instances for more than one AD converter, a unique
# configuration must be specified as argument for each instance.

#LOGLEVEL = logging.WARNING
LOGLEVEL = logging.WARNING

# 0 for main SPI bus, 1 for auxiliary SPI bus.
SPI_BUS = 0
# SPI clock rate in Hz. The ADS1256 supports a minimum of 1/10th of the output
# sample data rate in Hz to 1/4th of the oscillator CLKIN_FREQUENCY which
# results in a value of 1920000 Hz for the Waveshare board. However, since
# the Raspberry pi only supports power-of-two fractions of the 250MHz system
# clock, the closest value would be 1953125 Hz, which is slightly out of spec
# for the ADS1256. Choosing 250MHz/256 = 976563 Hz is a safe choice.
#SPI_FREQUENCY = 976563
SPI_FREQUENCY = 976563
# Risking the slightly out-of-spec speed:
#SPI_FREQUENCY = 1953125
# If set to True this will perform a chip reset using the hardware reset line
# when initializing the device.
CHIP_HARD_RESET_ON_START = False

#### Raspberry Pi GPIO configuration ##########################################
# =====> NEW in version 2 since using pigpio instead of wiringpi library:
# =====> Raspberry Pi pinning now uses the Broadcom numbering scheme!

# Tuple of all (chip select) GPIO numbers to be configured as an output and
# initialised to (inactive) logic high state before bus communication starts.
# Necessary for more than one SPI device if GPIOs are not otherwise handled.
#CHIP_SELECT_GPIOS_INITIALIZE = (22, 23)
CHIP_SELECT_GPIOS_INITIALIZE = (22, 23)
# Chip select GPIO pin number.
# This is required as hardware chip select can not be used with the ADS125x
# devices using this library
#CS_PIN      = 22
CS_PIN      = 22
# If DRDY is not connected to an input, a sufficient DRDY_TIMEOUT must be
# specified further below and aquisition will be slower.
#DRDY_PIN    = 17
DRDY_PIN    = 17
# Hardware reset pin is optional but strongly suggested in case multiple devices
# are connected to the bus as the ADS125x will lock-up in case multiple chips
# are selected simultaneously by accident.
#RESET_PIN   = 18 # Set to None if not used.
RESET_PIN   = 18 # Set to None if not used.
# Optional power down pin
#PDWN_PIN    = 27 # Set to None if not used.
PDWN_PIN    = 27 # Set to None if not used.
###############################################################################

##################  ADS1256 Constant Configuration Settings  ##################
# Seconds to wait in case DRDY pin is not connected or the chip
# does not respond. See table 21 of ADS1256 datasheet: When using a
# sample rate of 2.5 SPS and issuing a self calibration command,
# the timeout can be up to 1228 milliseconds:
DRDY_TIMEOUT    = 2
# Optional delay in seconds to avoid busy wait and reduce CPU load when
# polling the DRDY pin. Default is 0.000001 or 1 µs (timing not accurate)
DRDY_DELAY      = 0.000001
# Master clock rate in Hz. Default is 7680000:
CLKIN_FREQUENCY = 7680000
################################################################################


# All following settings are accessible through ADS1256 class properties

##############  ADS1256 Default Runtime Adjustable Properties  #################
# Analog reference voltage between VREFH and VREFN pins
v_ref = 2.5
# Gain seting of the integrated programmable amplifier. This value must be
# one of (GAIN_1, GAIN_2, GAIN_4, GAIN_8, GAIN_16, GAIN_32, GAIN_64).
# Gain = 1, V_ref = 2.5V ==> full-scale input voltage = 5.00V, corresponding
# to a 24-bit two's complement output value of 2**23 - 1 = 8388607
gain_flags = GAIN_1
################################################################################

####################  ADS1256 Default Register Settings  #######################
# REG_STATUS:
# When enabling the AUTOCAL flag: Any following operation that changes
# PGA GAIN, DRATE or BUFFER flags triggers a self calibration:
# THIS REQUIRES an additional timeout via WaitDRDY() after each such operation.
# Note: BUFFER_ENABLE means the ADC input voltage range is limited
# to (AVDD-2V),see datasheet
status = BUFFER_ENABLE
# REG_MUX:
# Default: positive input = AIN0, negative input = AINCOM
mux = POS_AIN0 | NEG_AINCOM
# REG_ADCON:
# Disable clk out signal (if not needed, source of disturbance),
# sensor detect current sources disabled, gain setting as defined above:
adcon = CLKOUT_OFF | SDCS_OFF | gain_flags
# REG_DRATE: 
# 10 SPS places a filter zero at 50 Hz and 60 Hz for line noise rejection
drate  = DRATE_10
# REG_IO: No GPIOs needed
gpio = 0x00
################################################################################


"""###########  Extended Description: ADS1256 Registers  ###############
Logically OR all desired option values together to form a 1 byte command
and write it to the respective register

Available Registers with address definitions:
        REG_STATUS  = 0x00
        REG_MUX     = 0x01
        REG_ADCON   = 0x02
        REG_DRATE   = 0x03
        REG_IO      = 0x04
        REG_OFC0    = 0x05
        REG_OFC1    = 0x06
        REG_OFC2    = 0x07
        REG_FSC0    = 0x08
        REG_FSC1    = 0x09
        REG_FSC2    = 0x0A


==> REG_STATUS, Status Register Configuration:

    Bits 7-4: ID3, ID2, ID1, ID0 Factory Programmed Identification Bits 
    (Read Only)

    Bit 3 ORDER: Data Output Bit Order

    0 = Most Significant Bit First (default)
    1 = Least Significant Bit First

    Input data is always shifted in most significant byte and bit first.
    Output data is always shifted out most significant byte first. The
    ORDER bit only controls the bit order of the output data within the
    byte.

    Bit 2 ACAL: Auto-Calibration

    0 = Auto-Calibration Disabled (default)
    1 = Auto-Calibration Enabled

    When Auto-Calibration is enabled, self-calibration begins at the
    completion of the WREG command that changes the PGA (bits 0-2 of ADCON
    register), DR (bits 7-0 in the DRATE register) or BUFEN (bit 1 in the
    STATUS register) values.

    THIS REQUIRES that an additional timeout (DREADY polling)
    is implemented after each such operation!


    Bit 1 BUFEN: Analog Input Buffer Enable

    0 = Buffer Disabled (default)
    1 = Buffer Enabled

    Bit 0 DRDY: Data Ready (Read Only)

    This bit duplicates the state of the DRDY pin, which is inverted logic.

    Option value definitions:
        BUFFER_ENABLE    = 0x02
        AUTOCAL_ENABLE   = 0x04
        ORDER_LSB        = 0x08


==> REG_MUX, Input multiplexer register, channel selection:

    High nibble selects positive input, low nibble negative input.

    Pin selection codes: logic OR together to form the register value.
    Example: ads1256.mux = POS_AIN0 | NEG_AINCOM

    # Pin selection codes for the positive input:
    POS_AIN0   = 0x00
    POS_AIN1   = 0x10
    POS_AIN2   = 0x20
    POS_AIN3   = 0x30
    POS_AIN4   = 0x40
    POS_AIN5   = 0x50
    POS_AIN6   = 0x60
    POS_AIN7   = 0x70
    POS_AINCOM = 0x80
    # Pin selection codes for the negative input:
    NEG_AIN0   = 0x00
    NEG_AIN1   = 0x01
    NEG_AIN2   = 0x02
    NEG_AIN3   = 0x03
    NEG_AIN4   = 0x04
    NEG_AIN5   = 0x05
    NEG_AIN6   = 0x06
    NEG_AIN7   = 0x07
    NEG_AINCOM = 0x08


==> REG_ADCON, A/D Control Register

    Bit 7 Reserved, always 0 (Read Only)

    Bits 6-5 CLK1, CLK0: D0/CLKOUT Clock Out Rate Setting

    00 = Clock Out OFF
    01 = Clock Out Frequency = fCLKIN (default)
    10 = Clock Out Frequency = fCLKIN/2
    11 = Clock Out Frequency = fCLKIN/4

    When not using CLKOUT, it is recommended that it be turned off. These
    bits can only be reset using the RESET pin.

    Bits 4-3 SDCS1, SCDS0: Sensor Detect Current Sources

    00 = Sensor Detect OFF (default)
    01 = Sensor Detect Current = 0.5uA
    10 = Sensor Detect Current = 2uA
    11 = Sensor Detect Current = 10uA

    The Sensor Detect Current Sources can be activated to verify the
    integrity of an external sensor supplying a signal to the ADS1255/6.
    A shorted sensor produces a very small signal while an open-circuit
    sensor produces a very large signal.

    Bits 2-0 PGA2, PGA1, PGA0: Programmable Gain Amplifier Setting

    Option value definitions:
        # Clock output pin setting
        CLKOUT_OFF      = 0x00
        CLKOUT_EQUAL    = 0x20
        CLKOUT_HALF     = 0x40
        CLKOUT_FOURTH   = 0x60

        # Sensor Detect Current Source
        SDCS_OFF     = 0x00
        SDCS_500pA   = 0x08
        SDCS_2uA     = 0x10
        SDCS_10uA    = 0x18

        # Gain level flag bits definitions:
        GAIN_1      = 0x00
        GAIN_2      = 0x01
        GAIN_4      = 0x02
        GAIN_8      = 0x03
        GAIN_16     = 0x04
        GAIN_32     = 0x05
        GAIN_64     = 0x06


==> DRATE Register: A/D Data Rate Address 0x03

    Bits 7-0 DR[7: 0]: Data Rate Setting(1)
    The 16 valid Data Rate settings. Only these are valid. 

    Sample rate definitions for CLKIN_FREQUENCY = 7.68 MHz (fCLKIN).
    Actual sample rates scale linearly with differing clock rates.

        DRATE_30000     = 0b11110000 # 30,000SPS (default)
        DRATE_15000     = 0b11100000 # 15,000SPS
        DRATE_7500      = 0b11010000 # 7,500SPS
        DRATE_3750      = 0b11000000 # 3,750SPS
        DRATE_2000      = 0b10110000 # 2,000SPS
        DRATE_1000      = 0b10100001 # 1,000SPS
        DRATE_500       = 0b10010010 # 500SPS
        DRATE_100       = 0b10000010 # 100SPS
        DRATE_60        = 0b01110010 # 60SPS
        DRATE_50        = 0b01100011 # 50SPS
        DRATE_30        = 0b01010011 # 30SPS
        DRATE_25        = 0b01000011 # 25SPS
        DRATE_15        = 0b00110011 # 15SPS
        DRATE_10        = 0b00100011 # 10SPS
        DRATE_5         = 0b00010011 # 5SPS
        DRATE_2_5       = 0b00000011 # 2.5SPS
"""


